
#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glext.h>
#include <math.h>
#include <stdio.h>
#include <assert.h>

#include "volume/GPUSlicer.h"
#include "mathutil/CVector.h"
#include "mathutil/CMatrix.h"
#include "core/GlutCgWindow.h"
#include "volume/Octree.h"
#include "volume/Block.h"
#include "app/globaldefs.h"

#define SQRT3 (1.7320508075f)

//#define ISOSURFACE
#define GRADIENT_ESTIMATION

PFNGLTEXIMAGE3DPROC glTexImage3D = NULL;

bool CGPUSlicer::s_bInit= false;

CGPUSlicer::CGPUSlicer()
{
	m_pOctree = NULL;
	m_pBlock = NULL;
	m_nDisplayListBase = 0;
	m_pListIndices = NULL;
	m_pTransferFunction = new CTransferFunction();
	m_pVolume = NULL;
	m_bShowOctree = false;
	m_bDisplayTF = false;
	m_bWireframe = false;
	m_nMin = 0;
	m_nMax = 255;
	m_dSliceDist = 0.002;

}


CGPUSlicer::~CGPUSlicer(void)
{
	if (NULL != m_pVolume) delete [] m_pVolume;
}

void CGPUSlicer::init() {

	if (s_bInit == false) {
		glTexImage3D             = (PFNGLTEXIMAGE3DPROC) wglGetProcAddress("glTexImage3D");
		assert(glTexImage3D != NULL);
		
		s_bInit = true;
	}


	m_pVertices[0] = CVector( -1.0, -1.0, -1.0,  1.0);
	m_pVertices[1] = CVector(  1.0, -1.0, -1.0,  1.0);
	m_pVertices[2] = CVector( -1.0,  1.0, -1.0,  1.0);
	m_pVertices[3] = CVector(  1.0,  1.0, -1.0,  1.0);

	m_pVertices[4] = CVector( -1.0, -1.0,  1.0,  1.0);
	m_pVertices[5] = CVector(  1.0, -1.0,  1.0,  1.0);
	m_pVertices[6] = CVector( -1.0,  1.0,  1.0,  1.0);
	m_pVertices[7] = CVector(  1.0,  1.0,  1.0,  1.0);

	loadVolume();

	initOctree(6,6,3);
	

}


void CGPUSlicer::initOctree(int sX, int sY, int sZ) {

	m_nMaxLevelX = sX;
	m_nMaxLevelY = sY;
	m_nMaxLevelZ = sZ;

	m_dScaleFactorX = 1.0/double(1 << (m_nMaxLevelX));
	m_dScaleFactorY = 1.0/double(1 << (m_nMaxLevelY));
	m_dScaleFactorZ = 1.0/double(1 << (m_nMaxLevelZ));

	if (NULL != m_pOctree) {
		delete m_pOctree;
	}

	m_pOctree = new COctree(m_nMaxLevelX,0,m_nMaxLevelY,0,m_nMaxLevelZ,0,-1.0,1.0,-1.0,1.0,-1.0,1.0,m_pVolume,XMAX,YMAX,ZMAX);

	initPrograms();
}


void
CGPUSlicer::render() {

	if (m_bWireframe) {
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
	} else {
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	}
	
	glMatrixMode(GL_MODELVIEW);
	glScaled(XSIZE,YSIZE,ZSIZE);

	CGprofile vertProfile = CGlutCgWindow::vertProfile();
	CGprofile fragProfile = CGlutCgWindow::fragProfile();
	cgGLDisableProfile(vertProfile);
	cgGLDisableProfile(fragProfile);

	static bool bInit = false;
	if (false == bInit) {
		loadPrograms();
		loadTextures();
		initDisplayLists();
		bInit = true;
	}

	float pMatrix[16];

	glGetFloatv(GL_MODELVIEW_MATRIX,pMatrix);
	glDisable(GL_LIGHTING);

	CMatrix matModelView(pMatrix[0],pMatrix[4],pMatrix[ 8],pMatrix[12],
						 pMatrix[1],pMatrix[5],pMatrix[ 9],pMatrix[13],
						 pMatrix[2],pMatrix[6],pMatrix[10],pMatrix[14],
						 pMatrix[3],pMatrix[7],pMatrix[11],pMatrix[15]);
	matModelView.invert();

	CVector camPosition = matModelView * CVector(0.0,0.0,0.0,1.0);

	CVector lightVec(0.0,0.0,1.0,0.0);
	CVector viewVec(-pMatrix[2],-pMatrix[6],-pMatrix[10],0.0);
	//CVector camPosition = pMatrix[14] * viewVec;
	camPosition[3] = 1.0;
	m_nMin = (VTYPE)m_pTransferFunction->getMin();
	m_nMax = (VTYPE)m_pTransferFunction->getMax();

	CVector scaledViewVec(viewVec[0]*m_dScaleFactorX,viewVec[1]*m_dScaleFactorY,viewVec[2]*m_dScaleFactorZ,0.0);

	double dMaxDist = scaledViewVec * m_pVertices[0];
	int nMaxIdx = 0;

	for(int i = 1; i < 8; ++i) {
		double dist = scaledViewVec * m_pVertices[i];
		if ( dist > dMaxDist) {
			dMaxDist = dist;
			nMaxIdx = i;
		}
	}

	double dXdist = 2.0*m_dScaleFactorX;
	double dYdist = 2.0*m_dScaleFactorY;
	double dZdist = 2.0*m_dScaleFactorZ;
	int nMaxNumSlices = int(sqrt(dXdist*dXdist + dYdist*dYdist + dZdist*dZdist)/m_dSliceDist)+1;
	
#if defined(ISOSURFACE)
	COctree::prepareForRendering(nMaxIdx, nMaxNumSlices, m_dSliceDist, m_nMin, VTYPE_MAX, camPosition, viewVec, m_vertexProgram);
#else
	COctree::prepareForRendering(nMaxIdx, nMaxNumSlices, m_dSliceDist, m_nMin, m_nMax, camPosition, viewVec, m_vertexProgram);
#endif

	if (m_bShowOctree) {
		glColor3f(1.0,0.0,0.0);
		m_pOctree->render(true);
	}

	cgGLEnableProfile(vertProfile);
	cgGLEnableProfile(fragProfile);
    cgGLBindProgram(m_vertexProgram);
	if (m_bWireframe) {
		cgGLBindProgram(m_fragmentProgramWF);
		cgGLSetTextureParameter( cgGetNamedParameter(m_fragmentProgramWF,"texture"),m_nTextureId);
		cgGLEnableTextureParameter(cgGetNamedParameter(m_fragmentProgramWF,"texture"));
		m_pTransferFunction->activate(m_fragmentProgramWF);
	} else {
		cgGLBindProgram(m_fragmentProgram);
		cgGLSetTextureParameter( cgGetNamedParameter(m_fragmentProgram,"texture"),m_nTextureId);
		cgGLEnableTextureParameter(cgGetNamedParameter(m_fragmentProgram,"texture"));
		m_pTransferFunction->activate(m_fragmentProgram);
	}

	// Bind uniform parameters to vertex shader
    cgGLSetStateMatrixParameter(cgGetNamedParameter(m_vertexProgram, "ModelViewProj"),
                                CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                CG_GL_MATRIX_IDENTITY);


#if defined(SHADING) | defined(ISOSURFACE) | defined(GRADIENT_ESTIMATION)
	cgGLSetParameter3dv(cgGetNamedParameter(m_fragmentProgram,"lightVec"),&(lightVec[0]));
	cgGLSetParameter3dv(cgGetNamedParameter(m_fragmentProgram,"viewVec"),&(viewVec[0]));
#endif

#if defined(ISOSURFACE) 
	glDisable(GL_BLEND);
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER,double(m_nMin)/255.0);
#else 
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#endif
	

	cgGLSetParameter3f(  cgGetNamedParameter(m_vertexProgram,"vecTranslate"),0.0,0.0,0.0);
	cgGLSetParameter3dv( cgGetNamedParameter(m_vertexProgram,"vecView"),&(viewVec[0]));
	cgGLSetParameter1f(  cgGetNamedParameter(m_vertexProgram,"frontIndex"),float(nMaxIdx));

	cgGLSetParameter1d(cgGetNamedParameter(m_vertexProgram,"dPlaneIncr"),m_dSliceDist);  
	m_pOctree->render();
	
	glDisable(GL_BLEND);


	
#if defined(ISOSURFACE) 
	glDisable(GL_ALPHA_TEST);
#else 
	glDisable(GL_BLEND);
#endif
	
	cgGLDisableProfile(vertProfile);
	cgGLDisableProfile(fragProfile);
	if (m_bDisplayTF) {
		m_pTransferFunction->render();
	}
}



void CGPUSlicer::loadPrograms()
{
	CGcontext context = CGlutCgWindow::context();
	CGprofile vertProfile = CGlutCgWindow::vertProfile();
	CGprofile fragProfile = CGlutCgWindow::fragProfile();
    assert(cgIsContext(context));

    // Load and compile the vertex program from demo_vert.cg; hold on to the
    // handle to it that is returned.
    m_vertexProgram = cgCreateProgramFromFile(context, CG_SOURCE, "..\\shader\\demo_vert_conditionals.cg",
                                            vertProfile, NULL, NULL);
    if (!cgIsProgramCompiled(m_vertexProgram))
        cgCompileProgram(m_vertexProgram);

    // Enable the appropriate vertex profile and load the vertex program.
    cgGLEnableProfile(vertProfile);
    cgGLLoadProgram(m_vertexProgram);


#ifdef SHADING
    // And similarly set things up for the fragment program.
    m_fragmentProgram = cgCreateProgramFromFile(context, CG_SOURCE, "..\\shader\\demo_frag_shading.cg",
                                              fragProfile, NULL, NULL);
    if (!cgIsProgramCompiled(m_fragmentProgram))
        cgCompileProgram(m_fragmentProgram);
#elif defined(ISOSURFACE)
    // And similarly set things up for the fragment program.
    m_fragmentProgram = cgCreateProgramFromFile(context, CG_SOURCE, "..\\shader\\demo_frag_isosurface.cg",
                                              fragProfile, NULL, NULL);
    if (!cgIsProgramCompiled(m_fragmentProgram))
        cgCompileProgram(m_fragmentProgram);
#elif defined(GRADIENT_ESTIMATION)
    // And similarly set things up for the fragment program.
    m_fragmentProgram = cgCreateProgramFromFile(context, CG_SOURCE, "..\\shader\\demo_frag_gradestimate.cg",
                                              fragProfile, NULL, NULL);
    if (!cgIsProgramCompiled(m_fragmentProgram))
        cgCompileProgram(m_fragmentProgram);
#else
    // And similarly set things up for the fragment program.
    m_fragmentProgram = cgCreateProgramFromFile(context, CG_SOURCE, "..\\shader\\demo_frag.cg",
                                              fragProfile, NULL, NULL);
    if (!cgIsProgramCompiled(m_fragmentProgram))
        cgCompileProgram(m_fragmentProgram);
#endif
	
    // And similarly set things up for the fragment program.
    m_fragmentProgramWF = cgCreateProgramFromFile(context, CG_SOURCE, "..\\shader\\demo_frag_wireframe.cg",
                                              fragProfile, NULL, NULL);
    if (!cgIsProgramCompiled(m_fragmentProgramWF))
        cgCompileProgram(m_fragmentProgramWF);

    cgGLEnableProfile(fragProfile);
    cgGLLoadProgram(m_fragmentProgramWF);
    cgGLLoadProgram(m_fragmentProgram);
	
	initPrograms();
}

void CGPUSlicer::initPrograms() {
	
	const float sequence[64] = {
		0.0f, 1.0f, 4.0f, 2.0f, 3.0f, 5.0f, 6.0f, 7.0f,      
		1.0f, 0.0f, 3.0f, 5.0f, 4.0f, 2.0f, 7.0f, 6.0f, 
		2.0f, 0.0f, 6.0f, 3.0f, 1.0f, 4.0f, 7.0f, 5.0f, 
		3.0f, 1.0f, 2.0f, 7.0f, 5.0f, 0.0f, 6.0f, 4.0f, 
		4.0f, 0.0f, 5.0f, 6.0f, 2.0f, 1.0f, 7.0f, 3.0f, 
		5.0f, 1.0f, 7.0f, 4.0f, 0.0f, 3.0f, 6.0f, 2.0f, 
		6.0f, 2.0f, 4.0f, 7.0f, 3.0f, 0.0f, 5.0f, 1.0f, 
		7.0f, 3.0f, 6.0f, 5.0f, 1.0f, 2.0f, 4.0f, 0.0f
	};

	const float e1[24] = {0.0f, 1.0f, 4.0f, 4.0f,   1.0f, 0.0f, 1.0f, 4.0f,   0.0f, 2.0f, 5.0f, 5.0f,   2.0f, 0.0f, 2.0f, 5.0f,   0.0f, 3.0f, 6.0f, 6.0f,   3.0f, 0.0f, 3.0f, 6.0f};
	const float e2[24] = {1.0f, 4.0f, 7.0f, 7.0f,   5.0f, 1.0f, 4.0f, 7.0f,   2.0f, 5.0f, 7.0f, 7.0f,   6.0f, 2.0f, 5.0f, 7.0f,   3.0f, 6.0f, 7.0f, 7.0f,   4.0f, 3.0f, 6.0f, 7.0f};
		
	float vertices[24];
	for(int i = 0; i < 8; ++i) {
		vertices[3*i]   = float(m_dScaleFactorX*m_pVertices[i][0]);
		vertices[3*i+1] = float(m_dScaleFactorY*m_pVertices[i][1]);
		vertices[3*i+2] = float(m_dScaleFactorZ*m_pVertices[i][2]);
	}

	cgGLSetParameterArray3f( cgGetNamedParameter(m_vertexProgram,"vecVertices"),  0,  0, vertices);
	cgGLSetParameterArray1f( cgGetNamedParameter(m_vertexProgram,"sequence"),     0,  0, sequence);
	cgGLSetParameterArray1f( cgGetNamedParameter(m_vertexProgram,"v1"),           0,  0, e1);
	cgGLSetParameterArray1f( cgGetNamedParameter(m_vertexProgram,"v2"),           0,  0, e2);

}



void CGPUSlicer::initDisplayLists() {
	
	m_nDisplayListBase = glGenLists(1);

	glNewList(m_nDisplayListBase,GL_COMPILE);
	
	glBegin(GL_TRIANGLE_FAN);
	glVertex2i(0,0);
	glVertex2i(1,0);
	glVertex2i(2,0);
	glVertex2i(3,0);
	glVertex2i(4,0);
	glVertex2i(5,0);
	glEnd();
	glEndList();

}

void CGPUSlicer::loadVolume() {

	FILE *pFile = fopen(FILENAME,"rb");
	if (NULL == pFile) {
		return exit(-1);
	}
	int size = XMAX*YMAX*ZMAX;
	m_pVolume = new unsigned char[size];
	memset(m_pVolume,0,size*sizeof(unsigned char));
	int ok = (int) fread(m_pVolume,sizeof(unsigned char), size,pFile);
	fclose(pFile);

}


#if defined(SHADING) || defined(ISOSURFACE)
void CGPUSlicer::loadTextures() {

	int size = XMAX*YMAX*ZMAX;
	unsigned char *pGradVolume = new unsigned char[size*4];
	
	int xidx1,xidx2;
	int yidx1,yidx2;
	int zidx1,zidx2;
	
	int idx=0;

	for(int z = 0; z < ZMAX; ++z) {
		zidx1 = (z == 0)? z: (z-1);
		zidx2 = (z == ZMAX-1)? z : (z+1);
		for(int y = 0; y < YMAX; ++y) {
			yidx1 = (y == 0)? y: (y-1);
			yidx2 = (y == YMAX-1)? y : (y+1);
			for(int x = 0; x < XMAX; ++x) {
				xidx1 = (x == 0)? x: (x-1);
				xidx2 = (x == XMAX-1)? x : (x+1);

				CVector gradient(
					double(m_pVolume[(z     * YMAX + y    ) * XMAX + xidx2]) - double(m_pVolume[(z     * YMAX + y    )*XMAX + xidx1]),
					double(m_pVolume[(z     * YMAX + yidx2) * XMAX + x    ]) - double(m_pVolume[(z     * YMAX + yidx1)*XMAX + x    ]),
					double(m_pVolume[(zidx2 * YMAX + y    ) * XMAX + x    ]) - double(m_pVolume[(zidx1 * YMAX + y    )*XMAX + x    ]),
					0.0);

				gradient.normalize();

				pGradVolume[idx++] = unsigned char(255.0*((gradient[0]+1.0)/2.0));
				pGradVolume[idx++] = unsigned char(255.0*((gradient[1]+1.0)/2.0));
				pGradVolume[idx++] = unsigned char(255.0*((gradient[2]+1.0)/2.0));
				pGradVolume[idx++] = m_pVolume[(z * YMAX + y)*XMAX+x];

			}
		}
	}

	glGenTextures(1,&m_nTextureId);

	glBindTexture(GL_TEXTURE_3D,m_nTextureId);

	// set the texture parameters
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage3D(GL_TEXTURE_3D,0,GL_RGBA,XMAX,YMAX,ZMAX,0,GL_RGBA,GL_UNSIGNED_BYTE,pGradVolume);

	delete [] m_pVolume;
	m_pVolume = NULL;
}

#else // !defined(SHADING)

void CGPUSlicer::loadTextures() {
	
	
	glGenTextures(1,&m_nTextureId);
	glBindTexture(GL_TEXTURE_3D,m_nTextureId);

	// set the texture parameters
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage3D(GL_TEXTURE_3D,0,GL_INTENSITY,XMAX,YMAX,ZMAX,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,m_pVolume);

	//delete [] m_pVolume;
	//m_pVolume = NULL;
}

#endif